<?php

namespace SimpleDBAL;


abstract class SQLBuilder
{

    private static function Quote($value, string $quote = "'")
    {
        if(strpos($value, "COUNT") !== false && strpos($value, "`") === false || $value == "NOW()") {
            $output = $value;
        } else {
            $output = (is_numeric($value) ? $value : ($value[0] == ":" && !strpos($value, "`") ? $value : $quote . $value . $quote));
        }
        return $output;
    }

    public static function Build(QueryBuilder $query)
    {
        switch ($query->getQType()) {
            case QueryBuilder::IS_SELECT:
                return self::GenerateSelect($query);
                break;
            case QueryBuilder::IS_UPDATE:
                return self::GenerateUpdate($query);
                break;
            case QueryBuilder::IS_INSERT:
                return self::GenerateInsert($query);
                break;
            case QueryBuilder::IS_TRUNCATE:
                return self::GenerateTruncate($query);
                break;
            case QueryBuilder::IS_DELETE:
                return self::GenerateDelete($query);
                break;

        }
    }


    /**
     * Generiert einen mit Kommata getrennten String
     * @param array $list
     * @param string $quote
     * @return string
     */
    private static function generateListItems(array $list, string $quote = "'")
    {
        $out = "";
        foreach ($list as $item)
            $out .= self::Quote($item, $quote) . ", ";

        return rtrim($out, ", ");
    }

    /**
     * Erzeugt eine Liste aus Schlüsseln und Werten
     * @param array $list
     * @param string $operator
     * @param string $seperator
     * @return string
     */
    private static function KeyValues(array $list, string $operator = " = ", string $seperator = ", ")
    {
        $out = "";
        foreach ($list as $key => $value)
            $out .= self::Quote($key, "`") . $operator . self::Quote($value, "'");

        return rtrim($out, $seperator);
    }

    /**
     * Erzeugt einen gültigen String für eine SELECT-Ausgabe eines SQL Statements
     * @param array $columns
     * @return string
     */
    private static function SelectConstruct(array $columns): string
    {
        return (count($columns) > 0) ? self::generateListItems($columns, "`") : "*";
    }

    private static function InsertConstruct(array $columns): string
    {
        return (count($columns) > 0) ? self::generateListItems($columns, "`") : "";
    }

    private static function ValueConstruct(array $values): string
    {
        return self::generateListItems($values);
    }

    private static function JoinConstruct(array $joins): string
    {
        $outJoin = "";
        foreach ($joins as $key => $value)
            $outJoin .= $key . " JOIN " . $value[0] . " ON `" . $value[0] . "`.`" . $value[1] . "` = `" . $value[2] . "`.`" . $value[3] . "`";

        return $outJoin;
    }

    /**
     * Erzeugt einen gültigen String für eine WHERE-Ausgabe eines SQL Statements
     * @param array $where
     * @return string
     */
    private static function WhereConstruct(array $where): string
    {
        if (count($where) === 0)
            return "";

        $out = "";
        foreach ($where as $item) {
            if($item->getOperator() == "BETWEEN") {
                $values = explode("#", $item->getValue());
                $out .= "`".$item->getColumn()."` ". $item->getOperator(). " " . self::Quote($values[0], "") ." AND ". self::Quote($values[1], "") ." AND ";
            }
            else {
                $out .= "`" . $item->getColumn() . "` " . $item->getOperator() . " " . self::Quote($item->getValue()) . " AND ";
            }
        }

        return rtrim($out, " AND ");
    }

    /**
     * Erzeugt das ORDER BY anhand eines Arrays das die Spaltennahmen als Array-Schlüssel und den boolischen Wert als Value enthält
     * True = Descending; False = Ascending;
     * @param array $orderBy
     * @return string
     */
    private static function OrderByConstruct(array $orderBy): string
    {
        if (count($orderBy) === 0)
            return "";

        $orderCols = [];
        foreach ($orderBy as $cols => $order)
            $orderCols[] = "`{$cols}`" . ($order === true ? " DESC" : "");

        return " ORDER BY " . self::generateListItems($orderCols, "");
    }

    /**
     * Erzeugt den GROUP BY Part des SQL Statements
     * @param $groupBy
     * @return string
     */
    private static function GroupByConstruct($groupBy): string
    {
        return " GROUP BY " . $groupBy;
    }

    /**
     * Erzeugt ein SELECT Statement welches mit JOINs, WHEREs, GROUP BYs und ORDER BYs erstellt wird
     * @param QueryBuilder $query
     * @return string
     */
    private static function GenerateSelect(QueryBuilder $query): string
    {
        return "SELECT " . self::SelectConstruct($query->getSelect()) . " FROM " . $query->getTable() .
            (!empty($query->getJoin()) ? " " . self::JoinConstruct($query->getJoin()) : "") .
            (!empty($query->getWhere()) ? " WHERE " . self::WhereConstruct($query->getWhere()) : "") .
            (!empty($query->getGroupBy()) ? self::GroupByConstruct($query->getGroupBy()) : "") .
            (!empty($query->getOrderBy()) ? self::OrderByConstruct($query->getOrderBy()) : "");
    }

    /**
     * Erzeugt ein UPDATE Statement
     * @param QueryBuilder $query
     * @return string
     */
    private static function GenerateUpdate(QueryBuilder $query): string
    {
        return "UPDATE " . $query->getTable() . " SET " . self::KeyValues($query->getSelect(), " = ", ",") .
            " WHERE " . self::WhereConstruct($query->getWhere());
    }

    /**
     * Erzeugt ein INSERT INTO Statement
     * @param QueryBuilder $query
     * @return string
     */
    private static function GenerateInsert(QueryBuilder $query): string
    {
        return "INSERT INTO " . $query->getTable() . " (" . self::InsertConstruct($query->getSelect()) . ") VALUES (" . self::ValueConstruct($query->getValues()) . ")";
    }

    /**
     * Erzeugt ein TRUNCATE TABLE Statement
     * @param QueryBuilder $query
     * @return string
     */
    private function GenerateTruncate(QueryBuilder $query): string
    {
        return "TRUNCATE TABLE ".$query->getTable();
    }

    /**
     * Erzeugt ein DELETE FROM Statement
     * @param QueryBuilder $query
     * @return string
     */
    private function GenerateDelete(QueryBuilder $query): string
    {
        return "DELETE FROM ". $query->getTable() . " WHERE " . self::WhereConstruct($query->getWhere());
    }


}